#pragma once

#include "File.h"

#include <filesystem>
#include <memory>
#include <unordered_map>
#include <iostream>
#include <locale>
#include <string>

using Path = std::filesystem::path;

class TomlTable
{
public:
	using Comment = std::pair<unsigned, std::string>;
	using KeyValuePairs = std::unordered_map<std::string, std::string>;

	TomlTable(const std::string& header)
		: mHeader(header)
	{

	}

	void addComment(const Comment& comment)
	{
		mComments.push_back(comment);
	}

	void addTable(std::unique_ptr<TomlTable> table)
	{
		mTables[table->getHeader()] = std::move(table);
	}

	void addKeyValuePair(const std::string& key, const std::string& value)
	{
		mMap[key] = value;
	}

	std::string getHeader() const
	{
		return mHeader;
	}

	TomlTable* getTable(const std::string& path)
	{
		return mTables[path].get();
	}

	KeyValuePairs getKeyValuePairs() const
	{
		return mMap;
	}

private:
	unsigned mLineOffset{ 0 };
	std::string mHeader;
	std::unordered_map<std::string, std::unique_ptr<TomlTable> > mTables;
	KeyValuePairs mMap;
	std::vector<Comment> mComments;
};

class TomlContent
{
public:
	TomlContent()
		: mRootTable(std::make_unique<TomlTable>("root"))
	{

	}

	TomlTable* getRootTable() const
	{
		return mRootTable.get();
	}

	TomlTable* getTable(const std::string& path) const
	{
		return mRootTable->getTable(path);
	}

private:
	std::unique_ptr<TomlTable> mRootTable;
};

class TomlReader
{
public:
	TomlReader()
		: mContent(std::make_unique<TomlContent>())
	{

	}

	TomlContent* getContent() const
	{
		return mContent.get();
	}

	void read(const Path& input_path)
	{
		const auto lines = File(input_path).readLines();
		mLastSectionOffset = 0;
		mWorkingTable = mContent->getRootTable();

		for (const auto& line : lines)
		{
			processLine(line);
			mLastSectionOffset++;
		}

		mWorkingTable = nullptr;
	}

	void processLine(const std::string& line)
	{
		bool in_comment{ false };
		bool in_header{ false };
		bool found_key{ false };
		std::string working_string;
		std::string key_string;
		for (auto c : line)
		{
			if (c == '#' && !in_comment)
			{
				in_comment = true;
			}
			else if (c == '[' && !in_comment && !in_header)
			{
				in_header = true;
			}
			else if (c == '=' && !in_comment && !in_header)
			{
				found_key = true;
				key_string = working_string;
				working_string = "";
			}
			else if (c == ']' && in_header)
			{
				break;
			}
			else
			{
				working_string += c;
			}
		}

		if (in_comment)
		{
			mWorkingTable->addComment({ mLastSectionOffset, working_string });
		}
		else if (in_header)
		{
			onHeader(working_string);
		}
		else if (found_key)
		{
			std::locale locale;
			key_string.erase(std::remove_if(key_string.begin(), key_string.end(), [locale](unsigned char c) {return std::isspace(c, locale); }), key_string.end());
			working_string.erase(std::remove_if(working_string.begin(), working_string.end(), [locale](unsigned char c) {return std::isspace(c, locale); }), working_string.end());
			onKeyValuePair(key_string, working_string);
		}
	}

	void onHeader(const std::string& header)
	{
		auto new_table = std::make_unique<TomlTable>(header);
		auto table_temp = new_table.get();
		mWorkingTable->addTable(std::move(new_table));
		mWorkingTable = table_temp;
	}

	void onKeyValuePair(const std::string key, const std::string value)
	{
		mWorkingTable->addKeyValuePair(key, value);
	}

private:
	unsigned mLastSectionOffset{ 0 };
	std::unique_ptr<TomlContent> mContent;
	TomlTable* mWorkingTable{nullptr};
};